-- Brainfuck interpreter
-- By Derek Newhall  (derek_newhall@yahoo.com)
-- Jan-2005
-- 
-- Reads in all of file before starting execution.


constant TRUE = 1, FALSE = 0   -- Truth values used in this program.

integer fn,          -- Number for file handler.
        loop_count,  -- Number of loop commands encountered,
                     --  should be 0 when finished parsing.
        index,       -- Pointer to current instruction in code.
        ptr          -- Pointer to current cell in array.

sequence file,       -- Command line options and, later, the file to be opened.
         code,       -- All the characters from file that are commands.
         array,      -- Array in memory that is manipulated.
         loop        -- Stack of

object line          -- Line read in from file. EOF is an integer.


-- Command line options
integer --Debug,      -- /D  -d : Turns debugger on.  REMOVED
        Winmode,    -- /W  -w : Turns on DOS/Windows newline support.
        Abort,      -- /A  -a : Aborts program if array pointer is negative.
        Commentchar -- /C  -c : Sets the (optional) line comment character.

-- Initialize all to FALSE.
Winmode = FALSE
Abort = FALSE
Commentchar = FALSE


-- Array manipulation constants
constant ARRAY_RESIZE = 10,    -- The amount of cells you want to add to
                               --  the array when it grows.
         INIT_ARRAY = ARRAY_RESIZE * 5, -- The initial size of the array.
         ARRAY_MAX = 0         -- The maximum value the array can grow to.
                               -- Only use if you absolutely need to limit
                               --  the size of the array. 0 is infinite.


-- Check to make sure the initial array is less than the maximum.
if ARRAY_MAX > 0 and ARRAY_MAX < INIT_ARRAY then
    puts(2, "Error: MAX_ARRAY constant must be bigger than INIT_ARRAY.\n")
    abort(1)
end if

-- Tests to see if a character is a valid command.
type valid_char(integer char)
    if char = '+' then
        return TRUE
    elsif char = '-' then
        return TRUE
    elsif char = '>' then
        return TRUE
    elsif char = '<' then
        return TRUE
    elsif char = '[' then
        return TRUE
    elsif char = ']' then
        return TRUE
    elsif char = ',' then
        return TRUE
    elsif char = '.' then
        return TRUE
    else
        return FALSE
    end if
end type

-- Get arguments off of command line and open file.
file = command_line()
if length(file) < 3 then
    puts(2, "Error: No source file specified.\n")
    abort(2)
else
    for i=3 to length(file) do
        -- If it starts with '/' or '-' then it's an option.
        if file[i][1] = '/' or file[i][1] = '-' and length(file[i]) >= 2 then

            -- Turn on Windows/DOS newline compatability.
            if file[i][2] = 'w' or file[i][2] = 'W' then
                Winmode = TRUE

            -- Turn on Windows/DOS newline compatability.
            elsif file[i][2] = 'a' or file[i][2] = 'A' then
                Abort = TRUE

            -- Sets the character after the 'c' to be the line comment character.
            elsif file[i][2] = 'c' or file[i][2] = 'C' then
                if length(file[i]) >= 3 then
                    Commentchar = file[i][3]
                else
                    Commentchar = '#'  -- ;/*'%#`
                end if

            -- Prints help info and exits.
            elsif file[i][2] = 'h' or file[i][2] = 'H'
              or file[i][2] = '?' then
                puts(1, "Brainfuck interpreter V1.0\tby Derek Newhall Jan-2005\n\n")
                if platform() = 3 then -- Linux/FreeBSD
                    puts(1, "If using Euphoria:\n eu bf.ex [-c[c]] [-w] [-a] filename\n\n")
                    puts(1, "If using an executable:\n bf [-c[c]] [-w] [-a] filename\n\n")
                    puts(1, "  filename\tSpecifies the file to be executed.\n")
                    puts(1, "  -c[c]\t\tTurns Line commenting with character c on.\n")
                    puts(1, "  \t\t  ('#' if not specified)\n")
                    puts(1, "  -w\t\tExecutes program with Windows/DOS newline support.\n")
                    puts(1, "  -a\t\tImmediately aborts program if array pointer moves left of the\n\t\t starting point.\n")
                else -- Windows/DOS
                    puts(1, "If using Euphoria:\n ex bf.ex [/C[c]] [/W] [/A] filename\n\n")
                    puts(1, "If using an executable:\n bf.exe [/C[c]] [/W] [/A] filename\n\n")
                    puts(1, "  filename\tSpecifies the file to be executed.\n")
                    puts(1, "  /C[c]\t\tTurns Line commenting with character c on.\n")
                    puts(1, "  \t\t  ('#' if not specified)\n")
                    puts(1, "  /W\t\tExecutes program with Windows/DOS newline support.\n")
                    puts(1, "  /A\t\tImmediately aborts program if array pointer moves left of the\n\t\t starting point.\n")
                end if
                abort(0)

            -- Unkown options
            else
                printf(2, "Error: Unknown option, %s.", {file[i]})
                abort(2)
            end if
        else
            -- If not an option then it's a file.
            file = file[i]
            exit
        end if
    end for
end if
-- Open file
fn = open(file, "r")
if fn = -1 then  -- If can't open file oor file doesn't exist.
    printf(2, "Error: Can't open file %s\n", {file})
    abort(2)
end if


code = {}              -- Initialize code sequence.
loop_count = 0         -- Counts the number of loop characters found.
-- Parse code for valid operator characters.
line = gets(fn)
while sequence(line) do   -- Do until line is an integer (EOF).
    if Commentchar = FALSE or line[1] != Commentchar then
        for i=1 to length(line) do
            if valid_char(line[i]) then
                code &= line[i]
                if line[i] = '[' then
                    loop_count += 1
                end if
                if line[i] = ']' then
                    loop_count -= 1
                    if loop_count < 0 then
                        puts(2, "Error: No matching '[' found.\n")
                        abort(2)
                    end if
                end if
            end if
        end for
    end if
    line = gets(fn)  -- Get next line.
end while
-- If loop_count is not 0 at the end of parsing then looping
--  charactars don't match up.
if loop_count > 0 then
    puts(2, "Error: No matching ']'.\n")
    abort(3)
end if



-- Default values
array = repeat(0, INIT_ARRAY)
loop = {}
ptr = 1

-- Execute the parsed code.
index = 1
while index <= length(code) do
    --  + : Increment pointer value operator.
    if code[index] = '+' then
        array[ptr] += 1

    --  - : Decrement pointer value operator.
    elsif code[index] = '-' then
        array[ptr] -= 1

    --  > : Move pointer to the right.
    elsif code[index] = '>' then
        ptr += 1
        if ptr > length(array) then
            array &= repeat(0, ARRAY_RESIZE)
            if ARRAY_MAX > 0 and length(array) > ARRAY_MAX then
                printf(2, "Error: Size of array exceeds limit of %d.\n", ARRAY_MAX)
                abort(7)
            end if
        end if

    --  < : Move pointer to the left.
    elsif code[index] = '<' then
        ptr -= 1
        if ptr < 1 then
            if Abort = TRUE then
                abort(0)
            else
                puts(2, "Error: Pointer moved outside of array.\n")
                abort(4)
            end if
        end if

    --  , : Get ASCII value of key pressed.
    elsif code[index] = ',' then
        array[ptr] = machine_func(26, 0)
        if Winmode = FALSE and array[ptr] = 13 then
            array[ptr] = 10
        end if

    --  . : Print ASCII character represented by the value of the pointer.
    elsif code[index] = '.' then
        printf(1, "%s", array[ptr])

    --  [ : Designates the start of a loop.
    elsif code[index] = '[' then
        loop &= index

    --  ] : If value under the pointer is not 0 then jump back to the start of the loop.
    elsif code[index] = ']' then
        if array[ptr] != 0 then
            -- This should NEVER happen.
            if length(loop) = 0 then
                puts(2, "Error: No matching ']'.\n  Please check parser.\n")
                abort(5)
            end if
            index = loop[length(loop)]  -- index = topmost value of loop
        else -- array[ptr] = 0
            loop = loop[1..length(loop)-1] -- Pop and discard topmost value from loop
        end if

   -- There should NEVER be an else.
    else
        puts(2, "Error: Bad operator.\n  Please check parser.\n")
        abort(5)
    end if

    index += 1  -- Increment index.
end while

